from abc import ABC, abstractmethod
import pandas as pd

class BaseEnsemble(ABC):
    """
    An interface for combining multiple models.
    """
    def __init__(self, models: list, returns_column, period, **kwargs):
        self.models = models
        self.returns_column = returns_column
        self.period = period
        self.asset_ids = set()
        self.features = set()
        for model in models:
            self.asset_ids.update(model.asset_ids)
            self.features.update(model.selected_features)

    @abstractmethod
    def publish(self, data):
        """
        Specify which data is passed to sink

        :param data: Panel data
        """
        pass

    @abstractmethod
    def realized_returns(self, data):
        """
        Specify how models are combined to generate returns

        :param data: Panel data
        """
        pass

    # average feature importance
    def feature_importances(self):
        """
        Combine features importances

        :return: feature importance
        """
        feature_importances = []
        for model in self.models:
            feature_importances.append(model.get_feature_importances())
        return pd.DataFrame(feature_importances).mean()

    # average shap values
    def shap_values(self, data):
        """
        Combine shap values

        :param data: Panel data
        :return: tuple (shap values, expected value, observations)
        """
        shap_values = []
        for model in self.models:  # average shap values
            shap_values_model, expected_value, observations = model.explain(data)
            shap_values.append(shap_values_model)
        shap_values = pd.concat(shap_values, axis=1, keys=range(len(self.models)))
        return shap_values.mean(axis=1, level=1), expected_value, observations

    def __len__(self):
        return len(self.models)



import pandas as pd
import quantstats as qs
from matplotlib import pyplot as plt
import matplotlib.ticker as mtick


def create_tearsheet(performance, benchmark_performance, file):
    """
    generate sophisticated performance analysis as html file
    """
    returns = performance.resample('1d').sum(min_count=1)
    returns.index = pd.to_datetime(returns.index.date, format="%Y-%m-%d")
    benchmark_returns = benchmark_performance.resample('1d').sum()
    benchmark_returns.index = pd.to_datetime(benchmark_returns.index.date, format="%Y-%m-%d")
    qs.reports.html(returns, benchmark_returns, compounded=False, periods_per_year=365, output=file)

def get_performance_metrics(performance, benchmark_performance, display=False):
    """
    get performance metrics
    """
    returns = performance.resample('1d').sum(min_count=1)
    returns.index = pd.to_datetime(returns.index.date, format="%Y-%m-%d")
    benchmark_returns = benchmark_performance.resample('1d').sum()
    benchmark_returns.index = pd.to_datetime(benchmark_returns.index.date, format="%Y-%m-%d")
    return qs.reports.metrics(returns, benchmark_returns, compounded=False, periods_per_year=365, prepare_returns=False, display=display)


# simple performance plot
def plot_backtest(performance, benchmark_performance):
    """
    plot backtest
    """
    ax = (performance * 100).cumsum().plot(title='Cumulative Returns', label='strategy')
    (benchmark_performance * 100).cumsum().plot(label='buy & hold', style='--')
    ax.yaxis.set_major_formatter(mtick.PercentFormatter())
    plt.legend()
    plt.show()

# get close and volume for multiple assets from binance
def get_crypto_data(client, symbols, since, until, interval):
    """
    get klines from binance for multiple symbols
    """
    expected_ts = pd.date_range(since, until, freq=interval, closed='right')
    asset_feature_columns = pd.MultiIndex.from_product([symbols, ['volume', 'return']])
    data = pd.DataFrame(index=expected_ts, columns=asset_feature_columns)
    for symbol in symbols:
        print(f'loading {symbol}')
        klines = client.get_historical_klines(symbol, interval, str(since), str(until))
        klines = pd.DataFrame(klines)
        klines.index = pd.to_datetime(klines[6], unit='ms', utc=True).round(interval)  # .asfreq(interval)
        data.loc[:, (symbol, 'volume')] = klines[5].astype('float')
        data.loc[:, (symbol, 'close')] = klines[4].astype('float')
    return data

# transform time series data to include lagged data (and percentage change)
def lagged_features(data: pd.DataFrame, period, n_periods) -> pd.DataFrame:
    """
    add lagged features to a DataFrame
    """
    assert n_periods > 10  # reasonable number of past periods is essential
    feature_names = data.fin.feature_names
    period_dt = pd.Timedelta(period)

    # extract changes and shifted values
    change = data.pct_change(freq=period_dt)\
        .rename(columns={name: f'{name}__change' for name in feature_names}, level=1)
    shift_1 = data.shift(freq=period_dt)\
        .rename(columns={name: f'{name}__shift1' for name in feature_names}, level=1)
    shift_2 = data.shift(freq=2 * period_dt)\
        .rename(columns={name: f'{name}__shift2' for name in feature_names}, level=1)

    # two non-overlapping periods for taking the mean
    midterm_periods = n_periods // 3  # one third
    longterm_periods = n_periods // 3 * 2  # two thirds
    rollingmean_midterm = data.rolling(midterm_periods * period_dt).mean().shift(freq=3 * period_dt)\
        .rename(columns={name: f'{name}__rollingmean_midterm' for name in feature_names}, level=1)
    rollingmean_longterm = data.rolling(longterm_periods * period_dt).mean().shift(freq=(midterm_periods + 3) * period_dt)\
        .rename(columns={name: f'{name}__rollingmean_longterm' for name in feature_names}, level=1)

    data_extended = pd.concat([change, shift_1, shift_2, rollingmean_midterm, rollingmean_longterm], axis=1)
    data_extended = data_extended.loc[data.index, :]
    return data_extended

def zscore(x, n_periods, period, min_periods=None):
    """
    rolling zscore
    """
    period_dt = pd.Timedelta(period)
    r = x.rolling(window=n_periods * period_dt, min_periods=min_periods)
    s = r.std(ddof=0).shift(freq=period_dt).reindex(x.index)
    m = r.mean().shift(freq=period_dt).reindex(x.index)
    z = (x-m)/s
    return z


import logging
import multiprocessing
import pickle
from typing import Type, List

import numpy as np
import optuna
import pandas as pd
from dateutil.tz import tzutc
from optuna.study import StudyDirection
from optuna.trial import FrozenTrial

from fintuna.ensemble.IndividualEnsemble import IndividualEnsemble
from fintuna.ensemble.BaseEnsemble import BaseEnsemble
from fintuna.model.ModelBase import ModelBase

n_cores = multiprocessing.cpu_count()
np.random.seed(0)
log = logging.getLogger('train')

class FinStudy:
    """A finstudy corresponds to a financial optimization task.
    A :class:`Model <fintuna.model.ModelBase>` is optimized based on the given `data`.
    It provides interfaces to get the out-of-sample performance and let's you fine-tune a model for production.
    """
    def __init__(self, Model: Type[ModelBase], data: pd.DataFrame, data_specs: dict = {}, split_specs: dict = {}, name: str = 'finstudy'):
        """
        :param Model:
        :param data: Panel data
        :param data_specs:
                Panel data specification.
                    'return_column' str (mandatory): column containing asset returns

                    'period' str (mandatory): aggregation period (e.g. '1h', '2d')

                    'sampling_freq' str (optional): sampling frequency (e.g. '1h', '2d'), defaults to period

                    'offset' str (optional): offset of sampling_freq (e.g. '20m')
        :param split_specs:
                Split specification for train, tune and eval data.
                    'tune_until' date string, int, float: end of tune data, defaults to 80% of the data

                    'train_until' date string, int, float: end of train data, defaults to 60% of tune data
        :param name: Name of study.
        """
        if 'sampling_freq' not in data_specs:
            if data.index.freqstr:
                self.sampling_freq = data.index.freqstr
            elif 'period' in data_specs:
                self.sampling_freq = data_specs['period']
            else:
                raise ValueError('Could not infer period from data. Please specify a period.')
        if 'period' not in data_specs:
            self.period = self.sampling_freq
        if 'period' in data_specs:
            self.period = data_specs['period']
        if 'sampling_freq' in data_specs:
            self.sampling_freq = data_specs['sampling_freq']

        if data.index.freqstr and pd.Timedelta(data.index.freqstr) != pd.Timedelta(self.sampling_freq):
            logging.warning(
                f'data.index.freq {data.index.freq} divers from sampling_freq {self.sampling_freq}. Using sampling_freq')

        if data.index.inferred_type != 'datetime64':
            raise ValueError('data index not supported. Please use datetime64 dataframe.')

        if not data.index.is_monotonic:
            raise ValueError('data index must be monotonic')
        if 'offset' not in data_specs:
            offsets = data.index - data.index.floor(self.sampling_freq)
            if len(offsets.unique()) == 1:
                self.offset = str(offsets[0]) if offsets[0].total_seconds else None
            else:
                raise ValueError('data has irregular offsets')
        else:
            self.offset = data_specs['offset']

        if 'tune_until' in split_specs:
            if type(split_specs['tune_until']) is str:
                self.tune_until = pd.Timestamp(split_specs['tune_until'], tz=tzutc())  # todo support multiple timezones
            if type(split_specs['tune_until']) is int:
                self.tune_until = data.index[split_specs['tune_until']]
            if type(split_specs['tune_until']) is float:
                self.tune_until = data.index[int(len(data) * split_specs['tune_until'])]
        else:
            self.tune_until = data.index[int(len(data) * .8)]

        test_len = len(data[:self.tune_until])
        if 'train_until' in split_specs:
            if type(split_specs['train_until']) is str:
                self.train_until = pd.Timestamp(split_specs['train_until'], tz=tzutc())
            if type(split_specs['train_until']) is int:
                self.train_until = data.index[split_specs['train_until']]
            if type(split_specs['train_until']) is float:
                self.train_until = data.index[int(test_len * split_specs['train_until'])]
        else:
            self.train_until = data.index[int(test_len * .6)]
        if self.train_until > self.tune_until:
            raise ValueError('train_until must be smaller than tune_until')

        self.returns_column = data_specs['returns_column']
        self.name = name
        self.data = data
        self.Model = Model

    def _create_ensemble(self, trials, data):
        models = []
        for trial in trials:
            model = self.Model(trial, **self.model_params)
            model.train(data, self.period, explainer=True)
            models.append(model)
        return self.ensemble_class(models, self.returns_column, self.period, **self.ensemble_params)

    # todo make private
    def _get_best_trials(self, n_trials) -> List[FrozenTrial]:
        trials_df = self.study.trials_dataframe()
        ascending = self.study.direction == StudyDirection.MINIMIZE
        best_trial_ids = trials_df.sort_values('value', ascending=ascending)[:n_trials]['number']
        best_trials = [trial for trial in self.study.get_trials() if trial.number in best_trial_ids]
        return best_trials

    def explore(self, ensemble_class: Type[BaseEnsemble] = IndividualEnsemble, study_params: dict = None, sampling_params: dict = None, model_params: dict = None, ensemble_size=1, ensemble_params: dict = None, conf_thrs_trials=20, n_trials=100) -> dict:
        """
        Train, Tune and explore the performance of a Model. Use Ensemble to combine the best performing models.

        :param ensemble_class: Ensemble to use
        :param study_params: Params of the `optuna study <https://optuna.readthedocs.io/en/v3.0.0/reference/generated/optuna.study.create_study.html#optuna.study.create_study>`_
        :param sampling_params:  Params of `optuna sampler <https://optuna.readthedocs.io/en/v3.0.0/reference/samplers/generated/optuna.samplers.TPESampler.html>`_
        :param model_params: Params of :py:class:`Model <fintuna.model.ModelBase>`
        :param ensemble_size: Number of best models to combine, defaults to 1 (= no ensemble)
        :param ensemble_params: Params of :class:`Model <fintuna.ensemble.BaseEnsemble>`
        :param conf_thrs_trials: Number of trials to simulate different confidence thresholds
        :param n_trials: Number of trials
        :return:
            Tuning results
                'feature_importances': pandas Series of feature importances

                'shap': tuple of (shap values, expected value, observations)

                'performance': out-of-sample returns (nan values indicate no exposure)

                'benchmark_performance': buy-and-hold returns weighted by exposure of 'performance'
        """
        study_params = study_params if study_params else {'direction': 'maximize'}
        if sampling_params:
            if 'n_startup_trials' in sampling_params:
                sampling_params['n_startup_trials'] *= conf_thrs_trials
        else:
            sampling_params = {
                'n_startup_trials': 10 * conf_thrs_trials,
                'seed': 0
                               }
        self.model_params = model_params if model_params else {}
        self.ensemble_params = ensemble_params if ensemble_params else {}
        self.ensemble_class = ensemble_class

        study_params['study_name'] = self.name

        self.ensemble_size = ensemble_size
        self.study = optuna.create_study(**study_params, sampler=optuna.samplers.TPESampler(**sampling_params))

        # train test split
        # exclude one prediction_dur to avoid look ahead bias at feature stacking
        explore_data_train = self.data[:self.train_until - pd.Timedelta(self.period)]
        explore_data_test = self.data[self.train_until:self.tune_until]

        log.info('start tuning')
        for i in range(n_trials):
            log.info(f'iteration {i + 1}/{n_trials}')
            tmp_trial = self.study.ask()
            model = self.Model(tmp_trial, **self.model_params)
            model.train(explore_data_train, self.period)

            predictions = model.predict(explore_data_test)
            returns = explore_data_test.loc[:, (model.asset_ids, [self.returns_column])]
            returns = returns - returns.xs(self.returns_column, axis=1, level=1).stack().mean()  # remove trend
            returns = returns.shift(freq=f'-{self.period}').reindex(predictions.index)
            
            # create multiple trials with varying confidence thresholds
            # note: there might be easier solutions as we are not sampling from conf_thrs (fintuna >= 0.0.4)
            min_prediction, max_prediction = predictions.stack().min(), predictions.stack().max()
            for thrs in np.random.uniform(min_prediction, max_prediction, conf_thrs_trials):
                realized_returns = model.realized_returns(predictions, thrs, returns, self.period)
                # todo make some checks on realized_returns

                performance = model.get_performance(realized_returns, self.period)

                # add trial copy to study
                params_thrs = tmp_trial.params.copy()
                params_thrs['conf_thrs'] = thrs
                dist_thrs = tmp_trial.distributions.copy()
                dist_thrs['conf_thrs'] = optuna.distributions.FloatDistribution(min_prediction, max_prediction)
                trial_thrs = optuna.trial.create_trial(params=params_thrs, distributions=dist_thrs, value=performance)
                self.study.add_trial(trial_thrs)
        # select best trials to form an ensemble
        best_trials = self._get_best_trials(self.ensemble_size)

        # held-out data
        self.real_data_train = self.data[:self.tune_until - pd.Timedelta(self.period)]
        self.real_data_test = self.data[self.tune_until:]
        self.ensemble = self._create_ensemble(best_trials, self.real_data_train)

        out_of_sample_realized_returns = self.ensemble.realized_returns(self.real_data_test)
        out_of_sample_realized_returns.name = 'performance'
        # calculate exposure
        exposed_dur = (~out_of_sample_realized_returns.isna()).sum() * pd.Timedelta(self.period)
        dur = out_of_sample_realized_returns.index[-1] - out_of_sample_realized_returns.index[0]
        exposure = exposed_dur / dur
        # get exposure adjusted benchmark
        benchmark_realized_returns = self.real_data_test.loc[:, (slice(None), self.returns_column)].resample(self.period).last().mean(axis=1) * exposure
        benchmark_realized_returns.name = 'benchmark_performance'

        feature_importances = self.ensemble.feature_importances()
        shap_values = self.ensemble.shap_values(self.real_data_test)
        # todo support asset specific explanations
        return {'feature_importances': feature_importances, 'shap': shap_values, 'performance': out_of_sample_realized_returns, 'benchmark': benchmark_realized_returns}


    def finish(self, n_candidates=None):
        """
        Prepares a model for production.
        Uses best models of tune data as candidates and sub-selects models that also perform well on eval data.
        Refits models on all data.

        :param n_candidates: Number of candidates to consider for selection, defaults to 5 * ensemble_size
        """
        if not n_candidates:
            n_candidates = self.ensemble_size * 5

        # todo store candidate models while exploring
        candidate_trials = self._get_best_trials(n_candidates)
        log.info('selecting best out-of-sample models')
        for trial in candidate_trials:
            model = self.Model(trial, **self.model_params)

            predictions = model.train(self.real_data_train, self.period).predict(self.real_data_test)
            returns_test = self.real_data_test.loc[:, (model.asset_ids, [self.returns_column])]
            returns_test = returns_test.shift(freq=f'-{self.period}').reindex(predictions.index)

            conf_thrs = model.trial.params['conf_thrs']
            real_returns = model.realized_returns(predictions, conf_thrs, returns_test, self.period)
            trial.performance = model.get_performance(real_returns, self.period)
            log.info(f'out-of-sample performance: {trial.performance}')
        final_trials = sorted(candidate_trials, key=lambda x: x.performance, reverse=True)[:self.ensemble_size]
    
        # refit on all data
        self.ensemble = self._create_ensemble(final_trials, self.data)
        self._data_validation = self.data[-10:]  # keep a tiny fraction of data for consistency checks
        self._pub_data_validation = [pd for pd in self.ensemble.publish(self._data_validation[-1:])]  # todo pick periods
        del self.data  # not needed anymore, makes pickeling cheaper
        del self.real_data_train
        del self.real_data_test




from fintuna.model.LongOnly import LongOnly
import numpy as np

class Pumps(LongOnly):
    """
    A long-only strategy focussing on abnormally high returns
    """
    def _init_classifier(self):
        clf = super()._init_classifier()
        clf.set_params(is_unbalance=True)  # use unbalanced sampling
        return clf

    def extract_label(self, data, period):
        """
        Predict if asset returns will be in the top quantile. The quantile is a hyperparameter ranging from 20% to 2%.
        """
        anomaly_magnitude = self.trial.suggest_float('anomaly_magnitude', .8, .98)
        returns = data.xs('return', axis=1, level=1)

        thrs = returns.stack().quantile(anomaly_magnitude)
        next_returns = returns.shift(freq=f'-{period}')
        labels = next_returns > thrs
        labels[next_returns.isna()] = np.nan
        return labels



import logging

import numpy as np
import pandas as pd
from apscheduler.schedulers.blocking import BlockingScheduler

from fintuna import FinStudy

logging.basicConfig(level = logging.INFO)
log = logging.getLogger('live')

def run(data_func, finstudy: FinStudy, sink):
    """
    This function calls a finstudy according to its sampling_freq periodically. It forwards the results to the sink.
    It also checks whether the data generating process (data_func) is valid (e.g. look-ahead-bias detection).

    :param data_func:
        A function which expects parameters (since, until, assets) and returns the panel data (see *Crypto End2End Prediction*)
    :param finstudy: The finstudy to deploy
    :param sink: A callable object that forwards or processes the output, e.g. a trading-engine (see :py:class:`fintuna.sink.LogSink`)
    """

    # check data_func
    since_val, until_val = finstudy._data_validation.index[0], finstudy._data_validation.index[-1]
    since_val -= pd.Timedelta(finstudy.period)  # data_func is right bound -> first observation at since + period
    data_validation_hat = data_func(since_val, until_val, finstudy._data_validation.fin.asset_names.to_list())
    pd.testing.assert_frame_equal(data_validation_hat, finstudy._data_validation)

    # test finstudy
    for i, pub_data in enumerate(finstudy.ensemble.publish(finstudy._data_validation[-1:])):
        for j, value in enumerate(pub_data):
            if type(value) is pd.Series:
                    np.testing.assert_allclose(value.values, finstudy._pub_data_validation[i][j].values)
            else:
                assert value == finstudy._pub_data_validation[i][j]
    log.info('checks complete!')

    scheduler = BlockingScheduler(timezone='utc')
    trigger_dur_dt = pd.Timedelta(finstudy.sampling_freq)

    now = pd.Timestamp.utcnow().floor(trigger_dur_dt) + pd.Timedelta(finstudy.offset)
    next_interval = now + trigger_dur_dt

    class LookaheadCheck:
        def __init__(self):
            self.prev_data = pd.DataFrame()

        def __call__(self, *args, **kwargs):
            now = pd.Timestamp.utcnow().floor(finstudy.sampling_freq)
            now += pd.Timedelta(finstudy.offset)
            # get data. also fetch the previous data point for lookahead checks
            since = now - pd.Timedelta(finstudy.period) - pd.Timedelta(finstudy.sampling_freq)
            data = data_func(since, now, finstudy.ensemble.asset_ids)
            # todo do some checks on the data

            cnt_data = data[-1:]
            prev_data_hat = data.iloc[0]  # watch for lookahead bias in the data
            try:
                _prev_data = self.prev_data.loc[prev_data_hat.name]
                identical = np.isclose(prev_data_hat, _prev_data, rtol=.01, equal_nan=True)
                if identical.sum() != len(_prev_data):
                    incorrect = pd.Series(~identical, index=_prev_data.index)
                    log.warning(f'Lookahead bias detected:\n {pd.concat([prev_data_hat.loc[incorrect], _prev_data.loc[incorrect]], axis=1)}')

                self.prev_data = self.prev_data.drop(_prev_data.name)
                self.prev_data = self.prev_data.append(data.iloc[-1])
            except KeyError as e:  # previous data is not available at startup
                log.debug('collecting data for lookahead checks..')
                self.prev_data = self.prev_data.append(data.iloc[-1])
                pass

            for args in finstudy.ensemble.publish(cnt_data):
                sink(*args)

    lookahead_check = LookaheadCheck()
    scheduler.add_job(lookahead_check, 'interval', seconds=int(trigger_dur_dt.total_seconds()),
                  start_date=next_interval.to_pydatetime())
    log.info(f'Scheduler started. {type(sink).__name__} will be triggered each {finstudy.sampling_freq}')
    scheduler.start()
